/*

GMP_int_ext.h -- extension to the GMP integer functions for the iRRAM library
 
Copyright (C) 2001 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

/*
Changelog: (initial version by Tom van Diessen)  

  2001-09-10 removal of several unneeded functions by Norbert
*/

#include <stdlib.h>
#include <stdio.h>
#include "gmp.h"
#ifdef __cplusplus
extern "C" {
#endif


#define GMP_min -1000000000
#define GMP_max 1000000000

/********** MP integer type definition **********/
/********** pointer to the MP integer type **********/

typedef mpz_ptr int_gmp_type;
typedef struct {unsigned long body;} int_gmp_sizetype;

/********** counting vars **********/

extern long int_gmp_var_count;
long int_gmp_getsize(int_gmp_type z);


/********** initialization function **********/

int_gmp_type int_gmp_init();

void int_gmp_free(int_gmp_type z);

/********** Conversion functions **********/

void int_gmp_int2int(long int i, int_gmp_type z);
void int_gmp_double2int(double d, int_gmp_type z);
void int_gmp_string2int(char* s, int_gmp_type z, int b);
//void int_gmp_real2int(ext_gmp_type r, int_gmp_type z);
long int_gmp_int2long(int_gmp_type z);

/********** standard arithmetic functions for MP integer **********/

void int_gmp_add(int_gmp_type z1, int_gmp_type z2, int_gmp_type z);
void int_gmp_sub(int_gmp_type z1, int_gmp_type z2, int_gmp_type z);
void int_gmp_mul(int_gmp_type z1, int_gmp_type z2, int_gmp_type z);
void int_gmp_div(int_gmp_type z1, int_gmp_type z2, int_gmp_type z);
void int_gmp_add_ui(int_gmp_type z1, unsigned long z2, int_gmp_type z);
void int_gmp_sub_ui(int_gmp_type z1, unsigned long z2, int_gmp_type z);
void int_gmp_mul_si(int_gmp_type z1,   signed long z2, int_gmp_type z);
void int_gmp_div_ui(int_gmp_type z1, unsigned long z2, int_gmp_type z);
void int_gmp_abs(int_gmp_type z1, int_gmp_type z);
void int_gmp_neg(int_gmp_type z1, int_gmp_type z);


/********** more MP integer functions *********/

void int_gmp_root(int_gmp_type z1, unsigned long int z2, int_gmp_type z);
void int_gmp_power_i(int_gmp_type z1, unsigned long int z2, int_gmp_type z);
void int_gmp_power_ii(unsigned long int z1, unsigned long int z2, int_gmp_type z);
void int_gmp_fac(unsigned long int z1, int_gmp_type z);
void int_gmp_modulo(int_gmp_type z1, int_gmp_type z2, int_gmp_type z3);
void int_gmp_sqrt(int_gmp_type z1, int_gmp_type z);
void int_gmp_shift(int_gmp_type z1, int_gmp_type z, long p);
long int_gmp_log(int_gmp_type z);


/********** comparison functions **********/

int int_gmp_cmp(int_gmp_type z1, int_gmp_type z2);
int int_gmp_sgn(int_gmp_type z);


/********** output functions for MP integers **********/

void int_gmp_writee(int_gmp_type z, long w);
char* int_gmp_swritee(int_gmp_type z, long w);
void int_gmp_write(int_gmp_type z, long w);
void int_gmp_outstr(int_gmp_type z, long w);
void int_gmp_printf(int_gmp_type z);
char* int_gmp_sprintf(int_gmp_type z);


/********** copying MP integers with/without initializing **********/

void int_gmp_duplicate_w_init(int_gmp_type z1, int_gmp_type *z2);
void int_gmp_duplicate_wo_init(int_gmp_type z1, int_gmp_type z2);

#ifdef __cplusplus
}
#endif

