/*

GMP_rat_ext.h -- extension to the GMP rational functions for the iRRAM library
 
Copyright (C) 2001 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


#include <stdlib.h>
#include <stdio.h>
#include "gmp.h"

#ifdef __cplusplus
extern "C" {
#endif


/********** MP rational type definition **********/
/********** pointer to the MP rational type **********/

typedef mpq_ptr rat_gmp_type;
//typedef struct {unsigned long body;} rat_gmp_sizetype;

/********** counting vars **********/

extern long rat_gmp_var_count;
int rat_gmp_getsize(rat_gmp_type z);
long rat_gmp_binsize(rat_gmp_type z);


/********** initialization function **********/

rat_gmp_type rat_gmp_init();

void rat_gmp_free(rat_gmp_type z);
long rat_gmp_size(rat_gmp_type z);
long rat_gmp_log(rat_gmp_type z);

void rat_gmp_canon(rat_gmp_type z);

// ********** standard arithmetic functions for MP rational **********
// ********** +-*/ *********

void rat_gmp_add(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_sub(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_mul(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_div(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_add_ui(rat_gmp_type z1, unsigned long z2, rat_gmp_type z);
void rat_gmp_sub_ui(rat_gmp_type z1, unsigned long z2, rat_gmp_type z);
void rat_gmp_mul_si(rat_gmp_type z1, unsigned long z2, rat_gmp_type z);
void rat_gmp_div_si(rat_gmp_type z1, unsigned long z2, rat_gmp_type z);
void rat_gmp_si_div(unsigned long z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_abs(rat_gmp_type z1, rat_gmp_type z);


// ********** more MP rational functions *********/
// ********** ^!%>>... **********/

void rat_gmp_power(rat_gmp_type z1, unsigned long int z2, rat_gmp_type z);
void rat_gmp_powerr(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z);
void rat_gmp_shift(rat_gmp_type z1, rat_gmp_type z, long p);


// ********** output functions for MP rationals **********

void rat_gmp_writee(rat_gmp_type z, long w);
char* rat_gmp_swritee(rat_gmp_type z, long w);
void rat_gmp_write(rat_gmp_type z, long w);
void rat_gmp_outstr(rat_gmp_type z, long w);
void rat_gmp_printf(rat_gmp_type z);
char* rat_gmp_sprintf(rat_gmp_type z);
void rat_gmp_string_2_rat(rat_gmp_type z, char* s);
void rat_gmp_double_2_rat(rat_gmp_type z, double d);

//void rat_gmp_get_numerator(int_gmp_type z, rat_gmp_type z1);
//void rat_gmp_get_denominator(int_gmp_type z, rat_gmp_type z1);


// ********** copying MP rationals with/without initializing **********

void rat_gmp_duplicate_w_init(rat_gmp_type z1, rat_gmp_type *z2);
void rat_gmp_duplicate_wo_init(rat_gmp_type z1, rat_gmp_type z2);

#ifdef __cplusplus
}
#endif

