/*

DYADIC.cc -- interface implementation of dyadic numbers for the iRRAM library
 
Copyright (C) 2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>

#include "iRRAM_core.h"

// Implementation of class DYADIC ********************
long DYADIC::precision=0;

DYADIC::DYADIC() {
   MP_init(value);
   MP_int_to_mp(0,value);
};

DYADIC::DYADIC(MP_type  y) {
  value=y;
// printf("TRANSFER(DYADIC):  %lu -> %lu\n",&y,&value);
//  MP_duplicate_w_init(y,value);
};

DYADIC::DYADIC(const long i) {
   MP_init(value);MP_int_to_mp(i,value);
};

DYADIC::DYADIC(const DYADIC& y) { 
  MP_duplicate_w_init(y.value,value);
};

DYADIC::~DYADIC() { 
  MP_clear(value);
};

DYADIC operator + (const DYADIC& x, const DYADIC& y) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_add(y.value,x.value,zvalue,DYADIC::precision);
  return zvalue;
};

DYADIC operator - (const DYADIC& x, const DYADIC& y) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_sub(y.value,x.value,zvalue,DYADIC::precision);
  return zvalue;
};

DYADIC operator - (const DYADIC& x) {
  DYADIC y;
  MP_type zvalue;
  MP_init(zvalue);
  MP_sub(y.value,x.value,zvalue,DYADIC::precision);
  return zvalue;
};



DYADIC operator * (const DYADIC& x, const DYADIC& y) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_mul(y.value,x.value,zvalue,DYADIC::precision);
  return zvalue;
};

DYADIC operator / (const DYADIC& x, const DYADIC& y) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_div(x.value,y.value,zvalue,DYADIC::precision);
  return zvalue;
};

#ifdef MP_has_shift
DYADIC scale(const DYADIC& x, long n) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_shift(x.value,zvalue,n);
  return DYADIC(zvalue);
}
#else
DYADIC scale(const DYADIC& x, long n) {
 exit(1);
//#error this can not be compiled
//#error the implementation is missing for this backend
}
#endif

#ifdef MP_double_to_mp 
DYADIC::DYADIC(const double d){
  MP_init(value); 
  MP_double_to_mp(d,value);
}
#else
DYADIC::DYADIC(const double d){
 exit(1);
//#error this can not be compiled
//#error the implementation is missing for this backend
}
#endif



DYADIC& DYADIC::operator = (const DYADIC& y) {
  MP_duplicate_wo_init(y.value,value);
  return (*this);
};

bool operator < (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg==-1);
};

bool operator <= (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg!=1);
};

bool operator > (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg==1);
};

bool operator >= (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg!=-1);
};

bool operator == (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg==0);
};

bool operator != (const DYADIC& x, const DYADIC& y) {
  long erg=MP_compare((x.value),(y.value));
  return  (erg!=0);
};

void rwrite (const DYADIC& x,const long w){
int width=w;
if (inlimit) {
    DEBUG1(0,"illegal output in continuous section!\n"); 
    return;
  }
  if (printfausgaben > printfaufrufe)  { 
    printfaufrufe++;
    } else  {
    printfausgaben++;
    printfaufrufe++;
    if (width<10) width=10;
    MP_writee(x.value,width);
    }
}

long size (const DYADIC& x) {
  return MP_size((x.value));
};

DYADIC abs (const DYADIC& x) {
  MP_type zvalue;
  MP_init(zvalue);
  MP_abs(x.value,zvalue);
  return zvalue;
};

