/*

REALS.cc -- main part of the implementation of reals for the iRRAM library
 
Copyright (C) 2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"


int infinite=0;

int debug=0;
int inlimit=0;
long printfausgaben=0;
long printfaufrufe=0;
char * eingabe;
char * position=eingabe;
int  eingabeplatz=150000;
long testanzahl=0;
long testaufrufe=0;
long * testwert;
long konvertanzahl=0;
long konvertaufrufe=0;
MP_type * konvert;

long starting_prec=-50;
long actual_prec=starting_prec;
long stiff_prec_inc=-50;
long prec_inc=0;
long prec_inc0=0;
long prec_inc1=-50;
long prec_inc2=-50;
double prec_factor=1.25;


REAL::REAL(MP_type y,const sizetype errorinfo): METRIC_OBJECT() {
  value=y;
  error=errorinfo;
};

REAL::REAL(): METRIC_OBJECT() {
  MP_init(value); 
  MP_int_to_mp(0,value); 
  sizetype_exact(error);
};

REAL::REAL(int i): METRIC_OBJECT()  {
  MP_init(value);
  MP_int_to_mp(i,value);
  sizetype_exact(error);
};


REAL::REAL(long i): METRIC_OBJECT()  {
  MP_init(value);
  MP_int_to_mp(i,value);
  sizetype_exact(error);
};

REAL::REAL(const DYADIC& y): METRIC_OBJECT()  { 
  MP_duplicate_w_init(y.value,value);
  sizetype_exact(error);
};

REAL::REAL(const REAL& y): METRIC_OBJECT()  {
  MP_duplicate_w_init(y.value,value);
  error=y.error;
};

REAL::REAL(char* s): METRIC_OBJECT()  {
  REAL result=atoREAL(s);
  error=result.error;
  MP_duplicate_w_init(result.value,value);
};

REAL::~REAL() { 
  if (value!=0) MP_clear(value);
  value=0;
};

REAL operator ^ (const REAL& x, const REAL& y) {
  return power(x,y);
}

REAL operator ^ (const REAL& x, long n) {
  return power(x,n);
}

REAL operator ^ (const REAL& x, int n) {
  return power(x,n);
}

REAL operator << (const REAL& x, long n) {
  return scale(x,n);
}

REAL operator >> (const REAL& x, long n) {
  return scale(x,-n);
}

REAL operator + (const REAL& x, const REAL& y) {
  MP_type zvalue; sizetype zerror; long local_prec;
  if (ACTUAL_STACK->prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,actual_prec);
  else {
  sizetype xsize,ysize;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  local_prec=max(xsize.exponent,ysize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_add(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator + (const REAL& x, long n)
{
  MP_type zvalue; sizetype zerror; long local_prec=0;
  if (ACTUAL_STACK->prec_policy == 0) 
    local_prec=max(x.error.exponent,actual_prec);
  else {
    sizetype xsize,ysize;
    MP_getsize(x.value,xsize);
    sizetype_set(ysize,n>0?n:-n,0);
    local_prec=max(xsize.exponent,ysize.exponent);
    local_prec=max(x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_addi(x.value,n,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
}

REAL operator + (long n, const REAL& x)
{
  return (x+n);
}

REAL operator + (int n, const REAL& x)
{
  return (x+long(n));
}

REAL operator + (const REAL& x,int n)
{
  return (x+long(n));
}

REAL& operator += (REAL& x,const REAL& y) {
  MP_type zvalue;
  sizetype zerror; long local_prec;
  if (ACTUAL_STACK->prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,actual_prec);
  else {
  sizetype xsize,ysize;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  local_prec=max(xsize.exponent,ysize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_add(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(x.error,y.error);
  sizetype_inc(x.error,zerror);
  ADJUST(x.error);
  MP_clear(x.value);
  x.value=zvalue;
  return x;
};

REAL operator - (const REAL& x, const REAL& y) {
  MP_type zvalue; sizetype zerror; long local_prec;
  if (ACTUAL_STACK->prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,actual_prec);
  else
  {
  sizetype xsize,ysize;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  local_prec=max(xsize.exponent,ysize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_sub(x.value,y.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator - (const REAL& x, long int n) 
{
  MP_type zvalue; 
  sizetype zerror; 
  long local_prec;
  if (ACTUAL_STACK->prec_policy == 0)
    local_prec=max(x.error.exponent,actual_prec);
  else {
    sizetype xsize,ysize;
    MP_getsize(x.value,xsize);
    sizetype_set(ysize,n>0?n:-n,0);
    local_prec=max(xsize.exponent,ysize.exponent);
    local_prec=max(x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_subi(x.value,n,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};


REAL operator - (long int n,const REAL& x) 
{
  	MP_type zvalue;
	sizetype zerror;
	long local_prec;
  	if (ACTUAL_STACK->prec_policy == 0)
  		local_prec=max(x.error.exponent,actual_prec);
  	else
  		{
  		sizetype xsize,ysize;
  		MP_getsize(x.value,xsize);
		sizetype_set(ysize,n>0?n:-n,0);
		local_prec=max(xsize.exponent,ysize.exponent);
  		local_prec=max(x.error.exponent,local_prec-50+actual_prec);
  		}
  	MP_init(zvalue);
  	MP_isub(n,x.value,zvalue,local_prec);
  	sizetype_set(zerror,1,local_prec);
  	sizetype_inc(zerror,x.error);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
};


REAL operator - (int n, const REAL& x)
{
  return (long(n) - x);
}

REAL operator - (const REAL& x,int n)
{
  return (x - long(n));
}




REAL operator - (const REAL& x) {
  REAL y=REAL(long(0));
  MP_type zvalue; sizetype zerror; long local_prec;
  if (ACTUAL_STACK->prec_policy == 0)
  local_prec=max3(y.error.exponent,x.error.exponent,actual_prec);
  else
  {
  sizetype xsize,ysize;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  local_prec=max(xsize.exponent,ysize.exponent);
  local_prec=max3(y.error.exponent,x.error.exponent,local_prec-50+actual_prec);
  }
  MP_init(zvalue);
  MP_sub(y.value,x.value,zvalue,local_prec);
  sizetype_set(zerror,1,local_prec);
  sizetype_inc(zerror,x.error);
  sizetype_inc(zerror,y.error);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator * (const REAL& x, const REAL& y) {
  MP_type zvalue;
  sizetype zerror,proderror,xsize,ysize;
  long local_prec;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  sizetype_mult(zerror,xsize,y.error);
  sizetype_mult(proderror,ysize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,y.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK->prec_policy == 0)
    local_prec=max(zerror.exponent,actual_prec);
  else
    local_prec=max(zerror.exponent,xsize.exponent+ysize.exponent-50+actual_prec);
  MP_init(zvalue);
  MP_mul(x.value,y.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL& operator *= (REAL& x,const REAL& y) {
  MP_type zvalue;
  sizetype zerror,proderror,xsize,ysize;
  long local_prec;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  sizetype_mult(zerror,xsize,y.error);
  sizetype_mult(proderror,ysize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,y.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK->prec_policy == 0)
    local_prec=max(zerror.exponent,actual_prec);
  else
    local_prec=max(zerror.exponent,xsize.exponent+ysize.exponent-50+actual_prec);
  MP_init(zvalue);
  MP_mul(x.value,y.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  x.error=zerror;
  MP_clear(x.value);
  x.value=zvalue;
  return x;
};

REAL operator * (const REAL& x, long int n) 
{
  	MP_type zvalue;
  	sizetype zerror,proderror,xsize,ysize;
  	long local_prec;
  	MP_getsize(x.value,xsize);
	sizetype_set(ysize,n>0?n:-n,0);
  	sizetype_mult(zerror,ysize,x.error);

  	if (ACTUAL_STACK->prec_policy == 0)
    		local_prec=max(zerror.exponent,actual_prec);
  	else
    		local_prec=max(zerror.exponent,xsize.exponent+ysize.exponent-50+actual_prec);
  	MP_init(zvalue);
  	MP_muli(x.value,n,zvalue,local_prec);
  	sizetype_set(proderror,1,local_prec);
  	sizetype_inc(zerror,proderror);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
}

REAL operator * (long n, const REAL& x){
  return (x*n);
}

REAL operator * (const REAL& x, int n) {
  return x*long(n);
}

REAL operator * (int n, const REAL& x) {
  return x*long(n);
}

REAL& operator *= (REAL& x,long int n) 
{	
  MP_type zvalue;
  sizetype zerror,proderror,xsize,ysize;
  long local_prec;
  MP_getsize(x.value,xsize);
  sizetype_set(ysize,n>0?n:-n,0);
  sizetype_mult(zerror,ysize,x.error);
  if (ACTUAL_STACK->prec_policy == 0)
    	local_prec=max(zerror.exponent,actual_prec);
  else
    	local_prec=max(zerror.exponent,xsize.exponent+ysize.exponent-50+actual_prec);
  MP_init(zvalue);
  MP_muli(x.value,n,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  x.error=zerror;
  MP_clear(x.value);
  x.value=zvalue;
  return x;

}

REAL& operator *= (REAL& x,int n) {
  return x*=long(n);
}

REAL operator / (const REAL& x, const REAL& y) {
  MP_type zvalue; 
  sizetype zerror,xsize,ysize,h1,h2,h3;
  long local_prec;
  MP_getsize(x.value,xsize);
  MP_getsize(y.value,ysize);
  sizetype_half(h1,ysize);
  if (sizetype_less(h1,y.error)) {
      DEBUG5(1,"insufficient precision %ld*2^(%ld) in denominator of size %ld*2^(%ld)\n",
              y.error.mantissa,y.error.exponent,ysize.mantissa,ysize.exponent);
      REITERATE(0);
    }
  sizetype_mult(h1,xsize,y.error);
  sizetype_mult(h2,ysize,x.error);
  sizetype_inc(h1,h2);
  h3=ysize;
  sizetype_dec(h3);
  sizetype_dec(h3,y.error);
  sizetype_mult(h2,h3,ysize);
  sizetype_div(zerror,h1,h2);
  if (ACTUAL_STACK->prec_policy == 0)
    local_prec=max(zerror.exponent,actual_prec);
  else
    local_prec=max(zerror.exponent,xsize.exponent-ysize.exponent-50+actual_prec);
  MP_init(zvalue);
  MP_div(x.value,y.value,zvalue,local_prec);
  sizetype_set(h2,1,local_prec);
  sizetype_inc(zerror,h2);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};

REAL operator / (const REAL& x, long int n) {
  	MP_type zvalue; 
  	sizetype zerror,xsize,ysize,h1;
  	long local_prec;
  	MP_getsize(x.value,xsize);
	sizetype_set(ysize,n>0?n:-n,0);
  	sizetype_div(zerror,x.error,ysize);
  	if (ACTUAL_STACK->prec_policy == 0)
    		local_prec=max(zerror.exponent,actual_prec);
  	else
    		local_prec=max(zerror.exponent,xsize.exponent-ysize.exponent-50+actual_prec);
  	MP_init(zvalue);
  	MP_divi(x.value,n,zvalue,local_prec);
  	sizetype_set(h1,1,local_prec);
  	sizetype_inc(zerror,h1);
  	ADJUST(zerror);
  	return REAL(zvalue,zerror);
}

REAL operator / (const REAL& x, int n) {
  return x/long(n);
}

REAL operator / (int n, const REAL& x) {
  return long(n)/x;
}


REAL operator / (long int n, const REAL& y) {
  	MP_type zvalue; 
  	sizetype zerror,xsize,ysize,h1,h2,h3,xerror;
  	long local_prec;
	sizetype_exact(xerror);
  	sizetype_set(xsize,n>0?n:-n,0);
  	MP_getsize(y.value,ysize);
  	sizetype_half(h1,ysize);

  	if (sizetype_less(h1,y.error)) 
		{
      	DEBUG5(1,"insufficient precision %ld*2^(%ld) in denominator of size %ld*2^(%ld)\n",y.error.mantissa,y.error.exponent,ysize.mantissa,ysize.exponent);
      	REITERATE(0);
    		}
  	sizetype_mult(h1,xsize,y.error);
  	sizetype_mult(h2,ysize,xerror);
  	sizetype_inc(h1,h2);
  	h3=ysize;
  	sizetype_dec(h3);
  	sizetype_dec(h3,y.error);
  	sizetype_mult(h2,h3,ysize);
  	sizetype_div(zerror,h1,h2);
  	if (ACTUAL_STACK->prec_policy == 0)
    		local_prec=max(zerror.exponent,actual_prec);
  	else
    		local_prec=max(zerror.exponent,xsize.exponent-ysize.exponent-50+actual_prec);
  	MP_init(zvalue);
  	MP_idiv(n,y.value,zvalue,local_prec);
  	sizetype_set(h2,1,local_prec);
  	sizetype_inc(zerror,h2);
  	ADJUST(zerror);

  	return REAL(zvalue,zerror);
}



REAL square (const REAL& x) {
  MP_type zvalue;
  sizetype zerror,proderror,xsize;
  long local_prec;
  MP_getsize(x.value,xsize);
  sizetype_mult(zerror,xsize,x.error);
  sizetype_mult(proderror,xsize,x.error);
  sizetype_inc(zerror,proderror);
  sizetype_mult(proderror,x.error,x.error);
  sizetype_inc(zerror,proderror);
  if (ACTUAL_STACK->prec_policy == 0)
    local_prec=max(zerror.exponent,actual_prec);
  else
    local_prec=max(zerror.exponent,xsize.exponent+xsize.exponent-50+actual_prec);
  MP_init(zvalue);
  MP_mul(x.value,x.value,zvalue,local_prec);
  sizetype_set(proderror,1,local_prec);
  sizetype_inc(zerror,proderror);
  ADJUST(zerror);
  return REAL(zvalue,zerror);
};


#ifdef MP_has_shift

REAL scale(const REAL& x, long n) {
  sizetype zerror;
  MP_type zvalue;
  MP_init(zvalue);
  MP_shift(x.value,zvalue,n);
  x.geterror(zerror);
  sizetype_shift(zerror,zerror,n);
  return REAL(zvalue,zerror);
}

#else

REAL scale(const REAL& x, long n) {
   if (n==0) return x;
   if (n==1) return 2*x;
   stiff_begin();
   REAL y=1;
   REAL xc=2;
   if (n<0) {xc=y/2;n=-n;}
   if (n==1) {
     stiff_end();
     return xc*x;
   }
   for (long k=n;k>0;k=k/2) { 
     if (k%2==1) y*=xc;
     if ( k ==1) break;
     xc*=xc;}
   stiff_end();
   return y*x; 
}

#endif


REAL& REAL::operator = (const REAL& y) {
  MP_duplicate_wo_init(y.value,value);
  error=y.error;
  return (*this);
};

LAZY_BOOLEAN operator < (const REAL& x, const REAL& y) {
  REAL z=y-x;
  sizetype s;
  MP_getsize(z.value,s);
  if (sizetype_less(s,z.error) )
	{
	DEBUG5(1,"insufficient precisions %ld*2^(%ld) and %ld*2^(%ld) in comparing\n",
              x.error.mantissa,x.error.exponent,
              y.error.mantissa,y.error.exponent);
        return BOTTOM;
	}
  return  ((MP_sign((z.value))==1));
};

LAZY_BOOLEAN operator <= (const REAL& x, const REAL& y) { 
  return  (x<y);
};

LAZY_BOOLEAN operator > (const REAL& x, const REAL& y) { 
  return  (y<x);
};

LAZY_BOOLEAN operator >= (const REAL& x, const REAL& y) { 
  return  (y<x);
};

LAZY_BOOLEAN operator == (const REAL& x, const REAL& y) { 
  return  (y<x)&&(x<y) ;
};

LAZY_BOOLEAN operator != (const REAL& x, const REAL& y) { 
  return  (y<x)||(x<y) ;
};

LAZY_BOOLEAN positive (const REAL& x, long k)
{ bool erg;
  sizetype xsize,ksize;
  sizetype_set(ksize,1,k);
  MP_getsize(x.value,xsize);
  if (sizetype_less(ksize,x.error) && sizetype_less(xsize,x.error) ) 
    {
      DEBUG3(1,"insufficient precision %ld*2^(%ld) in test on positive\n",
		x.error.mantissa,x.error.exponent);
      return BOTTOM;
    }
  erg=(MP_sign(x.value)==1);
  return erg; 
};

DYADIC approx (const REAL& x, const long p) {
  MP_type erg;
  sizetype psize;
  sizetype_set(psize,1,p+1);
  if ( ! inlimit) if (konvertanzahl > konvertaufrufe)
    { 
    MP_duplicate_w_init(konvert[konvertaufrufe],erg); 
    konvertaufrufe++;
    return DYADIC(erg);
    }
  if (sizetype_less(psize,x.error) )
     {
       DEBUG4(1,"insufficient precision %ld*2^(%ld) in approx(%ld)\n",
           x.error.mantissa,x.error.exponent,p);
       REITERATE(p-x.error.exponent);
     }
  MP_init(erg);
  MP_copy(x.value,erg,p-1);
  if ( ! inlimit) {
    MP_duplicate_w_init(erg,konvert[konvertaufrufe]);
    konvertanzahl++;
    konvertaufrufe++;
    }
  return DYADIC(erg); 
};

long size (const REAL& x) {
  sizetype xsize,ergsize;
  if (! inlimit) {
	if (testanzahl > testaufrufe)
	{ 
		testaufrufe++;
		return long(testwert[testaufrufe]);
		}
	}
  MP_getsize(x.value,xsize);
  sizetype_add(ergsize,xsize,x.error);
  sizetype_half(xsize,ergsize);
  sizetype_half(xsize,xsize);
  unsigned long result=0;
  unsigned long value =ergsize.mantissa;
  for (unsigned long m=32>>1; m>0; m=m>>1 ) {
    if (value>= (1<<m) ){
      result+=m; value=value>>m;
    }
  }
  if ( (1<<result) != ergsize.mantissa) result+=1;
  result+=ergsize.exponent;

  if (sizetype_less(xsize,x.error) )
	{  
	DEBUG5(1,"insufficient precision %ld*2^(%ld) in size %ld*2^(%ld)\n",
           x.error.mantissa,x.error.exponent,xsize.mantissa,xsize.exponent);
	REITERATE(0);
	}
  if ( ! inlimit)
	{
	testanzahl++;
	testaufrufe++;
        testwert[testaufrufe]=result;
	}
  return result;
};


#ifdef MP_double_to_mp 
REAL::REAL(double d): METRIC_OBJECT() {
  MP_init(value); 
  MP_double_to_mp(d,value);
  sizetype_exact(error);
}
#else
REAL::REAL(double d): METRIC_OBJECT() {
   REAL result =0L;
   long sign=0;
   long exp=0;
   long v;
   if ( d!= 0.0 ) {
     if  (d< 0) {sign=-1;d=-d;}
     else sign=1;
     while ( d >1024.0) {d=d/1024; exp+=10;}
     while ( d <1.0) {d=d*1024; exp-=10;}
     while ( d >0.0) {
       exp-=20;
       v=(long)(d*1048576L);
       result=result*1048576L+v;
       d=d*(1024*1024)-v;
     }
     if (exp < 0) {
       while ( exp<=-20 ) {exp+=20;result=result/(1048576L);}
       while ( exp<=-10 ) {exp+=10;result=result/(1024L);}
     } else {
       while ( exp>=20 ) {exp-=20;result=result*(1048576L);}
       while ( exp>=10 ) {exp-=10;result=result*(1024L);}
     }
   }
   result=sign*result;
   error=result.error;
   MP_duplicate_w_init(result.value,value);
}
#endif

REAL round2(const REAL& x) {
  MP_type erg;
  sizetype psize,error;
  sizetype_set(psize,1,-1);
  if ( ! inlimit) if (konvertanzahl > konvertaufrufe)
    { 
    MP_duplicate_w_init(konvert[konvertaufrufe],erg);
    konvertaufrufe++;
    sizetype_exact(error);
    return REAL(erg,error);
    }
  if (sizetype_less(psize,x.error) )
     {
       DEBUG3(1,"insufficient precision %ld*2^(%ld) in round\n",
           x.error.mantissa,x.error.exponent);
       REITERATE(0);
     }
  MP_init(erg);
  MP_truncate(x.value,erg);
  sizetype_exact(error);

  if ( ! inlimit) {
    MP_duplicate_w_init(erg,konvert[konvertaufrufe]);
    konvertanzahl++;
    konvertaufrufe++;
    }
  return REAL(erg,error); 
};

long round (const REAL& x) {
  sizetype xsize,cmpsize;
  long result;
  if (! inlimit) {
	if (testanzahl > testaufrufe)
	{ 
		testaufrufe++;
		return testwert[testaufrufe];
		}
	}
  MP_getsize(x.value,xsize);
  while  (xsize.exponent>-1) {
    xsize.mantissa=xsize.mantissa<<1;
    xsize.exponent-=1;
  }
  if (xsize.exponent<-8*sizeof(SIZETYPEMANTISSA))
     {xsize.exponent=-1; xsize.mantissa=0;}
  while  (xsize.exponent<-1) {
    xsize.mantissa=(xsize.mantissa>>1)+1;
    xsize.exponent+=1;
  }
  result=xsize.mantissa>>1;
  if(MP_sign(x.value)!=1) result=-result;
  sizetype_set(cmpsize,1,-2);
  if (sizetype_less(cmpsize,x.error) )
	{  
        MP_getsize(x.value,cmpsize);
	DEBUG5(1,"insufficient precision %ld*2^(%ld) in round %ld*2^(%ld)\n",
           xsize.mantissa,xsize.exponent,cmpsize.mantissa,cmpsize.exponent);
	REITERATE(0);
	}
  if ( ! inlimit)
	{
	testanzahl++;
	testaufrufe++;
	testwert[testaufrufe]=result;
	}
  return result;
};

long upperbound (const REAL& x) {
  sizetype xsize,ergsize;
  if (! inlimit) {
    if (testanzahl > testaufrufe)
    { 
    testaufrufe++;
    return testwert[testaufrufe];
    } }
  MP_getsize(x.value,xsize);
  sizetype_add(ergsize,xsize,x.error);
  while (ergsize.mantissa>1) {
    ergsize.mantissa=ergsize.mantissa>>1;
    ergsize.exponent+=1;
  }
  if ( ! inlimit) {
    testanzahl++;
    testaufrufe++;
    testwert[testaufrufe]=ergsize.exponent;
    }
  return ergsize.exponent;
};

LAZY_BOOLEAN bound (const REAL& x, const long k) {
  sizetype xsize,lowsize,ksize,highsize;
  long erg;
  sizetype_set(ksize,1,k);
  sizetype_set(lowsize,1,k-1);
  sizetype_inc(lowsize,x.error);
  MP_getsize(x.value,xsize);
  sizetype_add(highsize,xsize,x.error);
  if (sizetype_less(xsize,lowsize) && sizetype_less(ksize,highsize) ) 
    {  
      DEBUG6(1,"insufficient precision %ld*2^(%ld) in bounding by 2^(%ld) for argument of size  %ld*2^(%ld)\n",
              x.error.mantissa,x.error.exponent,k,xsize.mantissa,xsize.exponent);
      return BOTTOM;
    }
  erg=(sizetype_less(xsize,ksize) );
  return erg;
};

void rscanf (const char *format, void * input)
{
char zeile[1000];
int gelesen;

if (inlimit) {
    DEBUG1(0,"illegal input in continuous section!\n");
    exit(1);
  }
if(*position=='\0'){
  while (1) {
    if (position>eingabe+eingabeplatz-1000) {
      char* dummy=eingabe;
      eingabeplatz *= 2;
//   printf("%d %d %d\n",position,eingabe+eingabeplatz-1000,eingabeplatz);
      eingabe = new char[eingabeplatz];//***********************************
//   printf("%x %x %d\n",eingabe,dummy,position-dummy+1);
      memcpy(eingabe,dummy,eingabeplatz);
      position=eingabe+(position-dummy);
      delete dummy;
    }
    scanf("%s",zeile);
    strcat(position,zeile);
    strcat(position,"\n");
    gelesen=sscanf(position,format,input);
    if (1==gelesen) break;
    }
  position+=strlen(position)+1;
  *position='\0';
  }
  else
  {
   sscanf(position,format,input); 
   position+=strlen(position)+1;
  }
}

void stiff_begin (){ actual_prec += prec_inc;}
void stiff_end   (){ actual_prec -= prec_inc;}


void rprintf (const char *format, ...)
{
va_list paramlist;

if (inlimit) {
    DEBUG1(0,"illegal output in continuous section!\n");
    return;
  }
if (printfausgaben > printfaufrufe) 
  { printfaufrufe++;
    return;
   }
  printfausgaben++;
  printfaufrufe++;
  va_start(paramlist,format);
  vprintf(format,paramlist);

}

void   rwrite (const REAL& x, const long p, const long w)
{
sizetype psize;
long width=w;
if (inlimit) {
    DEBUG1(0,"illegal output in continuous section!\n"); 
    return;
  }
sizetype_set(psize,1,p);
if (sizetype_less(x.error,psize) ) {
  if (printfausgaben > printfaufrufe)  { 
    printfaufrufe++;
    } else  {
    printfausgaben++;
    printfaufrufe++;
    if (width<9) width=9;
    if (MP_size(x.value)<p) {
       printf("  0");
       for (int j=4;j<=width;j++) printf(" ");
       }
    else
       MP_writee(x.value,width);
    }
  } else {
	DEBUG4(1,"insufficient precision %ld*2^(%ld) in writing with precision 2^(%ld)\n",
		x.error.mantissa,x.error.exponent,p);
	REITERATE(p-x.error.exponent);
  }
}

void   rwrite (const REAL& x, const long w)
{
if (inlimit) {
    DEBUG1(0,"illegal output in continuous section!\n"); 
    return;
  }
sizetype psize;
long width=w;
if (width<9) width=9;
long p=-10*(width-8)/3;
sizetype_set(psize,1,p);
long s=MP_size(x.value);
long mantissa=(long)((s-x.error.exponent-GUARD_BITS)*.30103); 
if (sizetype_less(psize,x.error)|| (s>p && mantissa+8 < width) ) {
   DEBUG4(1,"insufficient precision %ld*2^(%ld) in writing with precision 2^(%ld)\n",
       x.error.mantissa,x.error.exponent,p);
   REITERATE(p-x.error.exponent);
   return;
  }

if (printfausgaben > printfaufrufe)  { 
    printfaufrufe++;
    return;
    }
        
printfausgaben++;
printfaufrufe++;

if (s<=p) {
  printf("  0");
  for (int j=4;j<=width;j++) printf(" ");
  return;
  }
MP_writee(x.value,width);
}

void   rwritee (const REAL& x, const long w)
{
if (inlimit) {
    DEBUG1(0,"illegal output in continuous section!\n"); 
    return;
  }
sizetype psize;
long width=w;
if (width<9) width=9;
long p=size(x)-2-10*(width-8)/3;
sizetype_set(psize,1,p);
long s=MP_size(x.value);
long mantissa=(long)((s-x.error.exponent-GUARD_BITS)*.30103); 
if (sizetype_less(psize,x.error)|| (s>p && mantissa+8 < width) ) {
   DEBUG4(1,"insufficient precision %ld*2^(%ld) in writing with precision 2^(%ld)\n",
       x.error.mantissa,x.error.exponent,p);
   REITERATE(p-x.error.exponent);
   return;
  }

if (printfausgaben > printfaufrufe)  { 
    printfaufrufe++;
    return;
    }
        
printfausgaben++;
printfaufrufe++;

MP_writee(x.value,width);
}

REAL abs (const REAL& x){
  MP_type zvalue;
  MP_init(zvalue);
  MP_abs(x.value,zvalue);
  return REAL(zvalue,x.error);
}

void rshow(const REAL& x, const long w)
{
if ( inlimit )
  {
  DEBUG1(0,"illegal output in continuous section!\n"); 
  return;
  }
if ( printfausgaben > printfaufrufe )  { 
  printfaufrufe++;
  return;
  }
  printfausgaben++;
  printfaufrufe++;
  int width=w;
  if ( width < 9 ) width =9;
  long mantissa=(long)((MP_size(x.value)-x.error.exponent-GUARD_BITS)*.30103); 
  if (mantissa < 1 ) { 
       printf(" .*");
       printf("E%05.0f",1+(x.error.exponent+GUARD_BITS)*.30103);
       for (long j=10;j<=width;j++) printf(" ");
       }
  else
  { 
    if ( mantissa>width-8 ) mantissa=width-8;
    MP_writee(x.value,mantissa+8);
    for (int j=mantissa+9;j<=width;j++) printf(" ");
  }
}

long module(REAL f(const REAL&),const REAL& x, long p){
  if (! inlimit) {
    if (testanzahl > testaufrufe)
    { 
    testaufrufe++;
    return testwert[testaufrufe];
    } }
  continous_begin();
  DYADIC d=approx(f(x),p);
  continous_end();
  sizetype ergsize=x.error;
  while (ergsize.mantissa>2) {
    ergsize.mantissa=ergsize.mantissa>>1;
    ergsize.exponent+=1;
  }
  if ( ! inlimit) {
    testanzahl++;
    testaufrufe++;
    testwert[testaufrufe]=ergsize.exponent;
    }
  return ergsize.exponent;
}


void REAL::rcheck() const
{
printf("Value: ");
MP_writee(value,30);
/*printf("\n"); MP_write(value,50);*/
printf(", Error: %ld*2^(%ld)\n\n",error.mantissa,error.exponent);
}



void continous_begin (){inlimit+=1;};
void continous_end   (){inlimit-=1;};

#define ABSOLUTE 0
#define REALTIVE 1

void precision_policy (long policy){ACTUAL_STACK->prec_policy=policy;};

void REAL::adderror (sizetype error)
{ sizetype_inc((*this).error,error);
  ADJUST((*this).error);
}

void REAL::seterror (sizetype error)
{ (*this).error=error;
}

void REAL::geterror (sizetype& error) const 
{ error=(*this).error;
}

/**************************************************************************/
extern "C" void iRRAM_initialize(int argc,char **argv)
{
int i;

for (i=0;i<argc;i+=1)
{
if (!strcmp(argv[i],"-d")) 
	{
	debug=1;
	DEBUG1(1,"Debugging Mode\n");
	}
if (!strncmp(argv[i],"-debug=",7)) 
	{
	debug=atoi(&(argv[i][7]));
	DEBUG2(1,"Debugging Level %d\n",debug);
	}
if (!strncmp(argv[i],"-prec_init=",11)) 
	{
	starting_prec=atoi(&(argv[i][11]));
	actual_prec=starting_prec;
	DEBUG2(1,"Initialising precision to 2^(%ld)\n",starting_prec);
	}
if (!strncmp(argv[i],"-prec_inc=",10)) 
	{
	int hi;
	hi=atoi(&(argv[i][10]));
	if (hi>0) prec_inc1=-hi;
	DEBUG2(1,"Initialising precision increment to %ld bits\n",-prec_inc1);
	}
if (!strncmp(argv[i],"-prec_factor=",13)) 
	{
	double hd;
	hd=atof(&(argv[i][13]));
	if (hd>1.0) prec_factor=hd;
	DEBUG2(1,"Initialising precision factor to %f\n",prec_factor);
	}
}
eingabe=new char[eingabeplatz];
testwert=new long[100000];
konvert= new MP_type[100000];
ikonvert=new MP_int_type[100000];
rkonvert=new MP_rat_type[100000];
MP_initialize;
}

