/*

convert.cc -- a few conversion functions for the iRRAM library
 
Copyright (C) 2002/2003 Norbert Mueller / Tom van Diessen
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"

REAL::REAL(const RATIONAL& r) : METRIC_OBJECT()
{
  INTEGER numi,deni;

  MP_rat_get_numerator(r.value,numi.value);
  MP_rat_get_denominator(r.value,deni.value);

  REAL result=REAL(numi)/REAL(deni);
  error=result.error;
  MP_duplicate_w_init(result.value,value);
}


//*****************************************************************************
// Constructing REAL from friend INTEGER
// Last change: 12.02.2001
//*****************************************************************************

REAL::REAL(const INTEGER& y): METRIC_OBJECT()
{
  MP_init(value);
  MP_INTEGER_to_mp(y.value,value);
  sizetype_exact(error);
}



//****************************************************************************************
// Constructing INTEGER from MP_object DYADIC
// Limited to GMP because of the limited availance of converting
// multiple precision floats to multiple precision integers
//****************************************************************************************


INTEGER::INTEGER(const DYADIC& y)
{
	MP_int_init(value);
	MP_mp_to_INTEGER(y.value,value);
}

//****************************************************************************************
// Constructing INTEGER from MP_object REAL
// see above
//****************************************************************************************
INTEGER::INTEGER(const REAL& x)
{
  	MP_int_type halb,one,temp2;
	MP_type temp;
	long p=5;
  	sizetype eins;
	MP_int_init(value);MP_int_init(one);
	MP_int_init(halb);MP_int_init(temp2);
	MP_int_to_INTEGER(5,halb);MP_int_to_INTEGER(1,one);
	MP_init(temp);
	sizetype_set(eins,3,-2);  // 3*2^(-2) = 0.75

  	if ( ! inlimit) if (ikonvertanzahl > ikonvertaufrufe)
    		{ 
		MP_int_duplicate_wo_init(ikonvert[ikonvertaufrufe],value);
    		ikonvertaufrufe++;
    		return;
    		}
	if (sizetype_less(eins,x.error))
		{			
       	DEBUG3(1,"insufficient precision %ld*2^(%ld) converting to integer\n",x.error.mantissa,x.error.exponent);
       	REITERATE(p-x.error.exponent);
		}
	MP_mp_to_INTEGER(x.value,value);//****** reelle Zahl abgeschnitten
/*	MP_INTEGER_to_mp(value,temp);
	MP_sub(x.value,temp,temp,p);
	MP_muli(temp,10,temp,p);
	MP_mp_to_INTEGER(temp,temp2);
	//printf("a\n");
	if (MP_int_compare(temp2,halb)>=0)
		MP_int_add(value,one,value);
	//printf("a2\n");
*/  	if ( ! inlimit) {
    		MP_int_duplicate_w_init(value,ikonvert[ikonvertaufrufe]);
    		ikonvertanzahl++;
    		ikonvertaufrufe++;
    		}
	//printf("a3\n");
}

//******************************************************************************************
// Constructing INTEGER from MP_object REAL including precision
// Added: 28.03.2001
// Last change: 03.04.2001
// REAL to INTEGER
//******************************************************************************************

INTEGER::INTEGER (const REAL& x, long p)
{
  	MP_int_type halb,one,temp2;
	MP_type temp;
  	sizetype psize,eins;
	MP_int_init(value);MP_int_init(one);
	MP_int_init(halb);MP_int_init(temp2);
	MP_int_to_INTEGER(5,halb);MP_int_to_INTEGER(1,one);
	MP_init(temp);
  	sizetype_set(psize,1,p-1);
  	if ( ! inlimit) if (ikonvertanzahl > ikonvertaufrufe)
    		{ 
		MP_int_duplicate_wo_init(ikonvert[ikonvertaufrufe],value);
    		ikonvertaufrufe++;
    		return;
    		}
  	if (sizetype_less(psize,x.error) )		//****** x.error>psize
     		{
       	DEBUG4(1,"insufficient precision %ld*2^(%ld) converting to integer (%ld)\n",x.error.mantissa,x.error.exponent,p);
       	REITERATE(p-x.error.exponent);
     		}
	MP_mp_to_INTEGER(x.value,value);		//****** reelle Zahl abgeschnitten
/*	MP_INTEGER_to_mp(value,temp);
	MP_sub(x.value,temp,temp,p);
	MP_muli(temp,10,temp,p);
	MP_mp_to_INTEGER(temp,temp2);
	if (MP_int_compare(temp2,halb)>=0)
		MP_int_add(value,one,value);
*/
  	if ( ! inlimit) 
		{
    		MP_int_duplicate_w_init(value,ikonvert[ikonvertaufrufe]);
    		ikonvertanzahl++;
    		ikonvertaufrufe++;
    		}
}

//Conversion from double to REAL

REAL::operator double const (){
  if ( bound(*this,-1150) ) return 0.0;
  long s=size(*this);
  DYADIC d=approx(*this,s-60);
  return MP_mp_to_double(d.value);
}

